//! This section deals with the memory manipulation functions. These should
//! be in the string module inside the rust module to more closely match
//! the 'C' library, but the no library feature needs them as well.
//! These functions are here so that there does not need to be
//! any duplication of the code.

// There is nothing to do on windows or macos as weak linkage seems to
// only work with ELF objects.
#![cfg_attr(all(feature="weak", not(windows), not(target_os="macos")),
            feature(linkage))]



/// Copies the values of n bytes from the location pointed to by
/// src directly to the memory block pointed to by dest.
///
/// The underlying type of the objects pointed to by both the src
/// and dest pointers are irrelevant for this function; The
/// result is a binary copy of the data.
///
/// The function does not check for any terminating null character
/// in source - it always copies exactly n bytes.
///
/// To avoid overflows, the size of the arrays pointed to by both
/// the dest and src parameters, shall be at least n bytes,
/// and should not overlap (for overlapping memory blocks, memmove
/// is a safer approach).
#[cfg_attr(all(feature="weak", not(windows), not(target_os="macos")),
           linkage="weak")]
#[no_mangle]
pub unsafe extern fn memcpy(dest: *mut u8, src: *const u8, n: usize)
   -> *mut u8
{
   let mut i: usize;

   i = 0;
   while i < n
   {
      *dest.offset(i as isize) = *src.offset(i as isize);
      i += 1;
   }

   return dest;
}

/// Copies the values of n bytes from the location pointed by src
/// to the memory block pointed by dest. Copying takes place
/// as if an intermediate buffer were used, allowing the dest
/// and src to overlap.
///
/// The underlying type of the objects pointed by both the src and
/// dest pointers are irrelevant for this function; The result
/// is a binary copy of the data.
///
/// The function does not check for any terminating null character
/// in source - it always copies exactly n bytes.
///
/// To avoid overflows, the size of the arrays pointed by both
/// the dest and src parameters, shall be at least n bytes.
#[cfg_attr(all(feature = "weak", not(windows), not(target_os = "macos")),
           linkage = "weak")]
#[no_mangle]
pub unsafe extern fn memmove(dest: *mut u8, src: *const u8, n: usize)
   -> *mut u8
{
   let mut i: usize;

   if src < dest as *const u8
   {
      // copy from end
      i = n;
      while i != 0
      {
         i -= 1;
         *dest.offset(i as isize) = *src.offset(i as isize);
      }
   }
   else
   {
      // copy from beginning
      i = 0;
      while i < n
      {
         *dest.offset(i as isize) = *src.offset(i as isize);
         i += 1;
      }
   }

   return dest;
}

/// Compares the first n bytes of the block of memory pointed by
/// s1 to the first num bytes pointed by s2, returning zero if
/// they all match or a value different from zero representing which
/// is greater if they do not.
///
/// Notice that, unlike strcmp, the function does not stop comparing
/// after finding a null character.
#[cfg_attr(all(feature = "weak", not(windows), not(target_os = "macos")),
           linkage = "weak")]
#[no_mangle]
pub unsafe extern fn memcmp(s1: *const u8, s2: *const u8, n: usize)
   -> i32
{
   let mut a: u8;
   let mut b: u8;
   let mut i: usize;

   i = 0;
   while i < n
   {
      a = *s1.offset(i as isize);
      b = *s2.offset(i as isize);

      if a != b
      {
         return (a as i32) - (b as i32)
      }

      i += 1;
   }

   return 0;
}

/// Sets the first n bytes of the block of memory pointed by s
/// to the specified value c (interpreted as an unsigned char).
#[cfg_attr(all(feature = "weak", not(windows), not(target_os = "macos")),
           linkage = "weak")]
#[no_mangle]
pub unsafe extern fn memset(s: *mut u8, c: i32, n: usize)
   -> *mut u8
{
   let mut i: usize;

   i = 0;
   while i < n
   {
      *s.offset(i as isize) = c as u8;
      i += 1;
   }

   return s;
}
